/* 
	Corsix's Image Retargetr
    Copyright (C) 2007 Peter Cawley

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Retarget.h"
#include <memory.h>
#include <math.h>
#include <queue>

#ifndef min
#define min(a,b) ( (a) < (b) ? (a) : (b) )
#endif

Retarget::Retarget()
{
	m_pOriginalImage = 0;
	m_pCurrentImage = 0;

	m_pPathfindCallback = 0;
	m_pPathfindCallbackToken = 0;
}

Retarget::~Retarget()
{
	delete m_pOriginalImage;
	delete m_pCurrentImage;
}

void Retarget::SetPathfindVisCallback(void (*pCallback)(unsigned char*, int, int, void*), void* pToken)
{
	m_pPathfindCallback = pCallback;
	m_pPathfindCallbackToken = pToken;
}

bool Retarget::Init(int iWidth, int iHeight, const unsigned char* pData)
{
	delete m_pOriginalImage;
	delete m_pCurrentImage;
	m_pCurrentImage = 0;
	m_pOriginalImage = new _image_t;

	m_pOriginalImage->iWidth = iWidth;
	m_pOriginalImage->iPitch = iWidth * 3;
	m_pOriginalImage->iHeight = iHeight;
	m_pOriginalImage->pData = new unsigned char[iWidth * iHeight * 3];
	m_pOriginalImage->pEnergyData = new long[iWidth * iHeight];
	m_pOriginalImage->pEnergyDataFlex = new long[iWidth * iHeight];

	if(m_pOriginalImage->pData == 0 || m_pOriginalImage->pEnergyData == 0)
		return false;

	if(pData == 0)
		memset(m_pOriginalImage->pData, 0, iWidth * iHeight * 3);
	else
		memcpy(m_pOriginalImage->pData, pData, iWidth * iHeight * 3);

	m_pOriginalImage->CalculateEnergyData();
	return true;
}

void Retarget::ReCalculateEnergyData()
{
	if(m_pCurrentImage)
	{
		m_pCurrentImage->CalculateEnergyData();
		if(m_pCurrentImage->pEnergyDataVisualised)
		{
			delete[] m_pCurrentImage->pEnergyDataVisualised;
			m_pCurrentImage->pEnergyDataVisualised = 0;
		}
	}
}

Retarget::_image_t::_image_t()
{
	pData = 0;
	pEnergyData = 0;
	pEnergyDataFlex = 0;
	iWidth = 0;
	iPitch = 0;
	iHeight = 0;
	pEnergyDataVisualised = 0;
	iAverageEnergy = 32;
}

Retarget::_image_t::_image_t(_image_t* pCopyFrom)
{
	iWidth = pCopyFrom->iWidth;
	iPitch = pCopyFrom->iPitch;
	iHeight = pCopyFrom->iHeight;
	iAverageEnergy = pCopyFrom->iAverageEnergy;

	int iSize = iWidth * iHeight;

	pData = new unsigned char[iSize * 3];
	pEnergyData = new long[iSize];
	pEnergyDataFlex = new long[iSize];
	pEnergyDataVisualised = 0;

	if(pData == 0 || pEnergyData == 0 || pEnergyDataFlex == 0)
		return;

	memcpy(pData, pCopyFrom->pData, iSize * 3);
	memcpy(pEnergyData, pCopyFrom->pEnergyData, iSize * sizeof(long));
	memcpy(pEnergyDataFlex, pCopyFrom->pEnergyDataFlex, iSize * sizeof(long));
}

Retarget::_image_t::~_image_t()
{
	delete[] pData;
	delete[] pEnergyData;
	delete[] pEnergyDataFlex;
	delete[] pEnergyDataVisualised;
}

void Retarget::_image_t::CalculateEnergyData()
{
	unsigned char* pSourcePixel = pData;
	long* pEnergyDataDest = pEnergyData;
	long* pFlexDest = pEnergyDataFlex;
	long iTotalEnergy = 0;
	for(int iY = 0; iY < iHeight; ++iY)
	{
		for(int iX = 0; iX < iWidth; ++iX)
		{
			long iDiffTotal = 0;
			long iDiffCount = 0;
			for(int iYSample = iY - 3; iYSample <= iY + 3; ++iYSample)
			{
				if(iYSample >= 0 && iYSample < iHeight)
				{
					for(int iXSample = iX - 3; iXSample <= iX + 3; ++iXSample)
					{
						if(iXSample >= 0 && iXSample < iWidth && (iXSample != iX || iYSample != iY) )
						{
							for(int iChannel = 0; iChannel < 3; ++iChannel)
							{
								iDiffTotal += abs(pSourcePixel[iChannel] - pData[iYSample * iPitch + (iXSample * 3) + iChannel]);
							}
							++iDiffCount;
						}
					}
				}
			}
			iDiffCount /= 2;
			pEnergyDataDest[0] = (iDiffTotal + iDiffCount / 2) / iDiffCount;
			pFlexDest[0] = 0;
			iTotalEnergy += pEnergyDataDest[0];

			pSourcePixel += 3;
			pEnergyDataDest += 1;
			pFlexDest += 1;
		}
	}

	iAverageEnergy = iTotalEnergy / (iWidth * iHeight * 2);
}

const unsigned char* Retarget::GetImageData() const
{
	_image_t* pImage = m_pCurrentImage ? m_pCurrentImage : m_pOriginalImage;
	return pImage->pData;
}

int Retarget::GetWidth() const
{
	_image_t* pImage = m_pCurrentImage ? m_pCurrentImage : m_pOriginalImage;
	return pImage->iWidth;
}

int Retarget::GetHeight() const
{
	_image_t* pImage = m_pCurrentImage ? m_pCurrentImage : m_pOriginalImage;
	return pImage->iHeight;
}

const unsigned char* Retarget::GetImageEnergyVisualisation()
{
	_image_t* pImage = m_pCurrentImage ? m_pCurrentImage : m_pOriginalImage;

	if(pImage->pEnergyDataVisualised == 0)
	{
		pImage->pEnergyDataVisualised = new unsigned char[pImage->iWidth * pImage->iHeight * 3];
		if(pImage->pEnergyDataVisualised == 0)
			return 0;

		long* pEnergyDataSrc = pImage->pEnergyData;
		long* pFlexSrc = pImage->pEnergyDataFlex;
		unsigned char* pVisDest = pImage->pEnergyDataVisualised;

		for(int iY = 0; iY < pImage->iHeight; ++iY)
		{
			for(int iX = 0; iX < pImage->iWidth; ++iX)
			{
				long iVal = pEnergyDataSrc[0] + pFlexSrc[0];
				iVal /= 3;

				if(iVal >= 0)
				{
					pVisDest[0] = 0;
					pVisDest[1] = (unsigned char) min( iVal, 255);
					pVisDest[2] = 0;
				}
				else
				{
					pVisDest[0] = (unsigned char) min(-iVal, 255);
					pVisDest[1] = 0;
					pVisDest[2] = 0;
				}

				pVisDest += 3;
				++pEnergyDataSrc;
				++pFlexSrc;
			}
		}
	}
	return pImage->pEnergyDataVisualised;
}

bool Retarget::RetargetDown(int iNewWidth, int iNewHeight)
{
	int iRecalculateTimer = 0;

	if(iNewWidth > m_pOriginalImage->iWidth || iNewHeight > m_pOriginalImage->iHeight)
		return false;

	if(m_pCurrentImage == 0 || iNewWidth > m_pCurrentImage->iWidth || iNewHeight > m_pCurrentImage->iHeight)
	{
		delete m_pCurrentImage;
		m_pCurrentImage = new _image_t(m_pOriginalImage);

		if(m_pCurrentImage->pData == 0 || m_pCurrentImage->pEnergyData == 0 || m_pCurrentImage->pEnergyDataFlex == 0)
			return false;
	}

	int iPathfindArrayLen = m_pCurrentImage->iWidth * m_pCurrentImage->iHeight;
	_pathfind_t *pPathfindArray = new _pathfind_t[iPathfindArrayLen];

	if(pPathfindArray == 0)
		return false;

	int *pColumnResult = new int[m_pCurrentImage->iHeight];
	int *pRowResult = new int[m_pCurrentImage->iWidth];

	if(pColumnResult == 0 || pRowResult == 0)
	{
		delete[] pRowResult;
		delete[] pColumnResult;
		delete[] pPathfindArray;
		return false;
	}

	while(m_pCurrentImage->iWidth > iNewWidth || m_pCurrentImage->iHeight > iNewHeight)
	{
		if(++iRecalculateTimer == 32)
		{
			iRecalculateTimer = 0;
			ReCalculateEnergyData();
		}

		bool bGotRow = false;
		bool bGotColumn = false;
		long iColumnScore, iRowScore;

		if(m_pCurrentImage->iWidth > iNewWidth)
		{
			for(int i = 0; i < iPathfindArrayLen; ++i)
				pPathfindArray[i].bInPath = false;

			bGotColumn = true;
			iColumnScore = _FindColumnToRemove(pPathfindArray, pColumnResult);
		}

		if(m_pCurrentImage->iHeight > iNewHeight)
		{
			for(int i = 0; i < iPathfindArrayLen; ++i)
				pPathfindArray[i].bInPath = false;

			bGotRow = true;
			iRowScore = _FindRowToRemove(pPathfindArray, pRowResult);
		}

		if(bGotRow && bGotColumn)
			bGotRow = iRowScore < iColumnScore;
		
		if(bGotRow)
			_RemoveRow(pRowResult);
		else
			_RemoveColumn(pColumnResult);
	}

	delete[] pRowResult;
	delete[] pColumnResult;
	delete[] pPathfindArray;

	ReCalculateEnergyData();

	return true;
}

bool Retarget::_pathfind_t::less::operator()(const _pathfind_t* a, const _pathfind_t* b)
{
	return (b->iCost + b->iHeuristic) < (a->iCost + a->iHeuristic);
}

long Retarget::_FindColumnToRemove(_pathfind_t* pPathfindArray, int* pResult)
{
	unsigned char* pVisData = 0;
	if(m_pPathfindCallback)
	{
		pVisData = new unsigned char[m_pCurrentImage->iWidth * m_pCurrentImage->iHeight * 3];
		memcpy(pVisData, m_pCurrentImage->pData, m_pCurrentImage->iWidth * m_pCurrentImage->iHeight * 3);
	}

	std::priority_queue<_pathfind_t*, std::vector<_pathfind_t*>, _pathfind_t::less> qToExplore;
	for(int iX = 0; iX < m_pCurrentImage->iWidth; ++iX)
	{
		pPathfindArray[iX].bInPath = true;
		pPathfindArray[iX].iCost = m_pCurrentImage->pEnergyData[iX] - m_pCurrentImage->pEnergyDataFlex[iX];
		pPathfindArray[iX].iHeuristic = 0;
		for(int iY = 1, iOffset = iX; iY < m_pCurrentImage->iHeight; ++iY, iOffset += m_pCurrentImage->iWidth)
			pPathfindArray[iX].iHeuristic += m_pCurrentImage->pEnergyData[iOffset] - m_pCurrentImage->pEnergyDataFlex[iOffset];

		pPathfindArray[iX].iFrom = iX;
		qToExplore.push(pPathfindArray + iX);

		if(pVisData)
		{
			pVisData[iX * 3] = 255;
			pVisData[iX * 3 + 1] = 0;
			pVisData[iX * 3 + 2] = 0;
		}
	}

	if(pVisData)
		m_pPathfindCallback(pVisData, m_pCurrentImage->iWidth, m_pCurrentImage->iHeight, m_pPathfindCallbackToken);

	_pathfind_t* pCurrent;
	int iTargetY = m_pCurrentImage->iHeight - 1;
	while(true)
	{
		pCurrent = qToExplore.top();
		int iY = (pCurrent - pPathfindArray) / m_pCurrentImage->iWidth;

		if(iY == iTargetY)
			break;

		int iX = (pCurrent - pPathfindArray) % m_pCurrentImage->iWidth;

		if(pVisData)
			pVisData[ iY * m_pCurrentImage->iPitch + iX * 3 +2] = 255;

		for(int iXNextRow = -1; iXNextRow <= +1; ++iXNextRow)
		{
			int iActualX = iX + iXNextRow;
			if( iActualX >= 0 && iActualX < m_pCurrentImage->iWidth)
			{
				int iOffset = (iY + 1) * m_pCurrentImage->iWidth + iActualX;
				_pathfind_t* pItem = pPathfindArray + iOffset;
				if(!pItem->bInPath)
				{
					pItem->bInPath = true;
					pItem->iFrom = iXNextRow;
					pItem->iCost = pCurrent->iCost + m_pCurrentImage->pEnergyData[iOffset] - m_pCurrentImage->pEnergyDataFlex[iOffset];
					pItem->iHeuristic = 0;
					for(int iY_ = iY + 2, iOffset = iX; iY_ < m_pCurrentImage->iHeight; ++iY_, iOffset += m_pCurrentImage->iWidth)
						pItem->iHeuristic += m_pCurrentImage->pEnergyData[iOffset] - m_pCurrentImage->pEnergyDataFlex[iOffset];

					qToExplore.push(pItem);

					if(pVisData)
					{
						pVisData[ (iY + 1) * m_pCurrentImage->iPitch + iActualX * 3 ] = 255;
						pVisData[ (iY + 1) * m_pCurrentImage->iPitch + iActualX * 3 +1] = 0;
						pVisData[ (iY + 1) * m_pCurrentImage->iPitch + iActualX * 3 +2] = 0;
					}
				}
			}
		}
		qToExplore.pop();

		if(pVisData)
			m_pPathfindCallback(pVisData, m_pCurrentImage->iWidth, m_pCurrentImage->iHeight, m_pPathfindCallbackToken);
	}

	if(pVisData)
		delete[] pVisData;

	long iFinalCost = pCurrent->iCost;

	for(; iTargetY >= 0; --iTargetY)
	{
		pResult[iTargetY] = pCurrent->iFrom;
		int iY = (pCurrent - pPathfindArray) / m_pCurrentImage->iWidth;
		int iX = (pCurrent - pPathfindArray) % m_pCurrentImage->iWidth;
		pCurrent = pPathfindArray + (iY - 1) * m_pCurrentImage->iWidth + iX - pCurrent->iFrom;
	}

	return iFinalCost;
}

void Retarget::_RemoveColumn(int* pColumn)
{
	int iRemovalX = 0;

	int iRemovedCount = 0;
	int iOffset = 0;

	if(m_pCurrentImage->pEnergyDataVisualised != 0)
	{
		delete[] m_pCurrentImage->pEnergyDataVisualised;
		m_pCurrentImage->pEnergyDataVisualised = 0;
	}

	for(int iY = 0; iY < m_pCurrentImage->iHeight; ++iY)
	{
		iRemovalX += pColumn[iY];
		for(int iX = 0; iX < m_pCurrentImage->iWidth; ++iX)
		{
			if(iX == iRemovalX)
			{
				++iRemovedCount;
			}
			else if(iRemovedCount > 0)
			{
				memcpy(m_pCurrentImage->pData + (iOffset - iRemovedCount) * 3, m_pCurrentImage->pData + iOffset * 3, 3);
				m_pCurrentImage->pEnergyData[iOffset - iRemovedCount] = m_pCurrentImage->pEnergyData[iOffset];
				m_pCurrentImage->pEnergyDataFlex[iOffset - iRemovedCount] = m_pCurrentImage->pEnergyDataFlex[iOffset];
			}

			++iOffset;
		}
	}

	--m_pCurrentImage->iWidth;
	m_pCurrentImage->iPitch -= 3;
}

long Retarget::_FindRowToRemove(_pathfind_t* pPathfindArray, int* pResult)
{
	unsigned char* pVisData = 0;
	if(m_pPathfindCallback)
	{
		pVisData = new unsigned char[m_pCurrentImage->iWidth * m_pCurrentImage->iHeight * 3];
		memcpy(pVisData, m_pCurrentImage->pData, m_pCurrentImage->iWidth * m_pCurrentImage->iHeight * 3);
	}

	std::priority_queue<_pathfind_t*, std::vector<_pathfind_t*>, _pathfind_t::less> qToExplore;
	for(int iY = 0; iY < m_pCurrentImage->iHeight; ++iY)
	{
		int iOffset = iY * m_pCurrentImage->iWidth;
		pPathfindArray[iOffset].bInPath = true;
		pPathfindArray[iOffset].iCost = m_pCurrentImage->pEnergyData[iOffset] - m_pCurrentImage->pEnergyDataFlex[iOffset];
		pPathfindArray[iOffset].iHeuristic = 0;
		for(int iX = 1; iX < m_pCurrentImage->iWidth; ++iX)
			pPathfindArray[iOffset].iHeuristic += m_pCurrentImage->pEnergyData[iOffset + iX] - m_pCurrentImage->pEnergyDataFlex[iOffset + iX];

		pPathfindArray[iOffset].iFrom = iY;
		qToExplore.push(pPathfindArray + iOffset);

		if(pVisData)
		{
			pVisData[iOffset * 3] = 255;
			pVisData[iOffset * 3 + 1] = 0;
			pVisData[iOffset * 3 + 2] = 0;
		}
	}

	if(pVisData)
		m_pPathfindCallback(pVisData, m_pCurrentImage->iWidth, m_pCurrentImage->iHeight, m_pPathfindCallbackToken);

	_pathfind_t* pCurrent;
	int iTargetX = m_pCurrentImage->iWidth - 1;
	while(true)
	{
		pCurrent = qToExplore.top();
		int iY = (pCurrent - pPathfindArray) / m_pCurrentImage->iWidth;
		int iX = (pCurrent - pPathfindArray) % m_pCurrentImage->iWidth;

		if(pVisData)
			pVisData[ iY * m_pCurrentImage->iPitch + iX * 3 +2] = 255;

		if(iX == iTargetX)
			break;

		for(int iYNextCol = -1; iYNextCol <= +1; ++iYNextCol)
		{
			int iActualY = iY + iYNextCol;
			if( iActualY >= 0 && iActualY < m_pCurrentImage->iHeight)
			{
				int iOffset = iActualY * m_pCurrentImage->iWidth + iX + 1;
				_pathfind_t* pItem = pPathfindArray + iOffset;
				if(!pItem->bInPath)
				{
					pItem->bInPath = true;
					pItem->iFrom = iYNextCol;
					pItem->iCost = pCurrent->iCost + m_pCurrentImage->pEnergyData[iOffset] - m_pCurrentImage->pEnergyDataFlex[iOffset];
					pItem->iHeuristic = 0;
					for(int iX_ = iX + 2; iX_ < m_pCurrentImage->iWidth; ++iX_)
						pItem->iHeuristic += m_pCurrentImage->pEnergyData[iOffset + iX_ -1] - m_pCurrentImage->pEnergyDataFlex[iOffset + iX_ -1];

					qToExplore.push(pItem);

					if(pVisData)
					{
						pVisData[ iActualY * m_pCurrentImage->iPitch + iX * 3 +3] = 255;
						pVisData[ iActualY * m_pCurrentImage->iPitch + iX * 3 +4] = 0;
						pVisData[ iActualY * m_pCurrentImage->iPitch + iX * 3 +5] = 0;
					}
				}
			}
		}
		qToExplore.pop();

		if(pVisData)
			m_pPathfindCallback(pVisData, m_pCurrentImage->iWidth, m_pCurrentImage->iHeight, m_pPathfindCallbackToken);
	}

	if(pVisData)
		delete[] pVisData;

	long iFinalCost = pCurrent->iCost;

	for(; iTargetX >= 0; --iTargetX)
	{
		pResult[iTargetX] = pCurrent->iFrom;
		int iY = (pCurrent - pPathfindArray) / m_pCurrentImage->iWidth;
		int iX = (pCurrent - pPathfindArray) % m_pCurrentImage->iWidth;
		pCurrent = pPathfindArray + (iY - pCurrent->iFrom) * m_pCurrentImage->iWidth + iX - 1;
	}

	return iFinalCost;
}

void Retarget::_RemoveRow(int* pRow)
{
	if(m_pCurrentImage->pEnergyDataVisualised != 0)
	{
		delete[] m_pCurrentImage->pEnergyDataVisualised;
		m_pCurrentImage->pEnergyDataVisualised = 0;
	}

	int iRemovalY = 0;
	for(int iX = 0; iX < m_pCurrentImage->iWidth; ++iX)
	{
		iRemovalY += pRow[iX];
		for(int iY = iRemovalY + 1; iY < m_pCurrentImage->iHeight; ++iY)
		{
			int iOffset = iY * m_pCurrentImage->iWidth + iX;

			memcpy(m_pCurrentImage->pData + iOffset * 3 - m_pCurrentImage->iPitch, m_pCurrentImage->pData + iOffset * 3, 3);
			m_pCurrentImage->pEnergyData[iOffset - m_pCurrentImage->iWidth] = m_pCurrentImage->pEnergyData[iOffset];
			m_pCurrentImage->pEnergyDataFlex[iOffset - m_pCurrentImage->iWidth] = m_pCurrentImage->pEnergyDataFlex[iOffset];
		}
	}

	--m_pCurrentImage->iHeight;
}